@@ -12,8 +12,34 @@ from keydnn.conv_utils import ConvTools
 
 import numpy as np 
 
+class V1NetworkLayer(NetworkLayer):
+        
+    def __init__(self, *args, **kwargs) -> None:
+        pass
+
+    def clear_gradients(self) -> None:
+        pass
+
+    def optimize(self, *args, **kwargs) -> None:
+        pass
+
+    def get_weights(self) -> None:
+        return None
+
+    def without_gradients(self) -> None:
+        pass
+
+    def use_gradients(self) -> None:
+        pass
+
+    def forward(self, *args, **kwargs) -> None:
+        raise NotImplementedError("The forward function has not been implemented.")
+
+    def backward(self, *args, **kwargs) -> None:
+        raise NotImplementedError("The backward function has not been implemented.")
+
 @NeuralNetwork.register_layer
-class Linear(NetworkLayer):
+class Linear(V1NetworkLayer):
 
     @classmethod
     def export_to_dict(cls, layer : "Linear") -> Dict[ str, Union[ str, int, bytes ] ]:
@@ -68,6 +94,8 @@ class Linear(NetworkLayer):
     def __init__(self, input_size  : int,
                        output_size : int,
                        activation  : Optional[ Union[ Activation, str ] ] = None) -> None:
+        
+        super(Linear, self).__init__()
 
         if (self.DEBUG_MODE):
 
@@ -100,7 +128,7 @@ class Linear(NetworkLayer):
 
         self._W = initialize_weights((self._output_size, self._input_size), self._input_size)
 
-        self._B = initialize_weights((self._output_size, 1), self._input_size)
+        self._B = np.zeros(shape = (self._output_size, 1)) #initialize_weights((self._output_size, 1), self._input_size)
 
         # optimizer parameters : temporary values and gradients
 
@@ -211,21 +239,32 @@ class Linear(NetworkLayer):
         if (any((x is None) for x in [ self._X, self._Z, self._A ])):
             raise ValueError("Back propagation before forward propagation is forbidden. Try calling `forward` first.")
 
-        # (B, O) : initialize ones-matrix if activation is undefined
-        dA_dZ = np.ones_like(gradient)
-
-        if (self._activation is not None):
-
-            # (B, O) : partial derivative of activation with respect to Z
-            dA_dZ = self._activation.call_derivative(self._Z)
-
         # Z Formula : [ Z = WX + B : (O, 1) = (O, I) x (I, 1) ]
 
         # (O, 1) : initialized with ones
         dZ_dB = np.ones_like(self._B)
 
+
+        # (B, O) : initialize ones-matrix if activation is undefined
+        # dA_dZ = np.ones_like(gradient)
+
         # (B, O) <= (B, O) * (B, O)
-        dZ = gradient * dA_dZ
+        dZ = gradient
+
+        if (self._activation is not None):
+
+            # (B, O) : partial derivative of activation with respect to Z
+            # dA_dZ = self._activation.call_derivative(self._Z)
+
+            # (B, O) : (B, O) * (B, O) or (B, O) x (B, O, O)
+            # dZ = self._activation.call_derivative(self._Z, gradient)
+
+            # (B, O)
+            dZ = self._activation.apply_gradient(self._Z, gradient)
+
+        # (B, O) <= (B, O) * (B, O)
+        # dZ = gradient * dA_dZ
+
 
         # (O, 1) <= (O,) <= (B, O) <= (B, O) x (1, O)
         dB = np.reshape(np.sum(dZ * dZ_dB.T, axis = 0), (self._output_size, 1))
@@ -245,7 +284,7 @@ class Linear(NetworkLayer):
         return dX
 
 @NeuralNetwork.register_layer
-class Conv2D(NetworkLayer):
+class Conv2D(V1NetworkLayer):
 
     @classmethod
     def export_to_dict(cls, layer : "Conv2D") -> Dict[ str, Union[ str, int, bytes ] ]:
@@ -316,6 +355,8 @@ class Conv2D(NetworkLayer):
                        
                 ) -> None:
         
+        super(Conv2D, self).__init__()
+        
         if (self.DEBUG_MODE):
 
             assert isinstance(input_channels, int)
@@ -388,7 +429,7 @@ class Conv2D(NetworkLayer):
         self._W = initialize_weights((self._output_channels, self._input_channels, self._kernel_height, self._kernel_width), N)
 
         # (CC, 1)
-        self._B = initialize_weights((self._output_channels, 1), N)
+        self._B = np.zeros(shape = (self._output_channels, 1)) #initialize_weights((self._output_channels, 1), N)
 
         # optimizer parameters : temporary values and gradients
 
@@ -517,18 +558,23 @@ class Conv2D(NetworkLayer):
             raise ValueError("Back propagation before forward propagation is forbidden. Try calling `forward` first.")
 
         # (B, CC, OH, OW) : initialize ones-matrix if activation is undefined
-        dA_dZ = np.ones_like(gradient)
+        # dA_dZ = np.ones_like(gradient)
+
+        dZ = gradient
 
         if (self._activation is not None):
 
             # (B, CC, OH, OW) : partial derivative of activation with respect to Z
-            dA_dZ = self._activation.call_derivative(self._Z)
+            # dA_dZ = self._activation.call_derivative(self._Z)
+
+            # (B, CC, OH, OW)
+            dZ = self._activation.apply_gradient(self._Z, gradient)
 
         # (CC, 1) : initialized with ones
         dZ_dB = np.ones_like(self._B)
 
         # (B, CC, OH, OW) <= (B, CC, OH, OW) * (B, CC, OH, OW)
-        dZ = gradient * dA_dZ
+        # dZ = gradient * dA_dZ
 
         # (CC, 1) <= (B, CC, OH, OW) x (CC, 1)
         dB = np.reshape(np.einsum("BXHW,XO->XO", dZ, dZ_dB, optimize = True), (self._output_channels, 1))
@@ -554,7 +600,7 @@ class Conv2D(NetworkLayer):
         return dX
 
 @NeuralNetwork.register_layer
-class Flatten(NetworkLayer):
+class Flatten(V1NetworkLayer):
 
     @classmethod
     def export_to_dict(cls, layer : "Flatten") -> Dict[ str, Union[ str, int] ]:
@@ -590,6 +636,8 @@ class Flatten(NetworkLayer):
                        input_height   : int,
                        input_width    : int) -> None:
         
+        super(Flatten, self).__init__()
+        
         if (self.DEBUG_MODE):
 
             assert isinstance(input_height, int)
@@ -648,7 +696,7 @@ class Flatten(NetworkLayer):
         return gradient.reshape((-1, self._input_channels, self._input_height, self._input_width))
 
 @NeuralNetwork.register_layer
-class LazyFlatten(NetworkLayer):
+class LazyFlatten(V1NetworkLayer):
 
     @classmethod
     def export_to_dict(cls, layer : "LazyFlatten") -> Dict[ str, Union[ str, int] ]:
@@ -671,22 +719,7 @@ class LazyFlatten(NetworkLayer):
         return flatten_layer
         
     def __init__(self) -> None:
-        pass
-
-    def clear_gradients(self) -> None:
-        pass
-
-    def optimize(self, *args, **kwargs) -> None:
-        pass
-
-    def get_weights(self) -> None:
-        return None
-
-    def without_gradients(self) -> None:
-        pass
-
-    def use_gradients(self) -> None:
-        pass
+        super(LazyFlatten, self).__init__()
 
     def forward(self, x : np.ndarray) -> np.ndarray:
 
@@ -708,7 +741,7 @@ class LazyFlatten(NetworkLayer):
         return gradient.reshape((self._B, self._C, self._H, self._W))
 
 @NeuralNetwork.register_layer
-class MaxPooling2D(NetworkLayer):
+class MaxPooling2D(V1NetworkLayer):
 
     @classmethod
     def export_to_dict(cls, layer : "MaxPooling2D") -> Dict[ str, Union[ str, int] ]:
@@ -740,6 +773,8 @@ class MaxPooling2D(NetworkLayer):
     def __init__(self, kernel_size : Tuple[ int, int ], *,
                        num_workers : Optional[ int ] = 4) -> None:
         
+        super(MaxPooling2D, self).__init__()
+        
         if (self.DEBUG_MODE):
 
             assert isinstance(kernel_size, tuple)
@@ -754,21 +789,6 @@ class MaxPooling2D(NetworkLayer):
 
         self._num_workers = num_workers
 
-    def clear_gradients(self) -> None:
-        pass
-
-    def optimize(self, *args, **kwargs) -> None:
-        pass
-
-    def get_weights(self) -> None:
-        return None
-
-    def without_gradients(self) -> None:
-        pass
-
-    def use_gradients(self) -> None:
-        pass
-
     def forward(self, x : np.ndarray) -> np.ndarray:
 
         # x : (B, C, H, W)
@@ -803,6 +823,70 @@ class MaxPooling2D(NetworkLayer):
 
         return gradient
 
+@NeuralNetwork.register_layer
+class Dropout(V1NetworkLayer):
+
+    @classmethod
+    def export_to_dict(cls, layer : "Dropout") -> Dict[ str, Union[ str, int] ]:
+
+        if (cls.DEBUG_MODE):
+            assert isinstance(layer, Dropout)
+
+        output_data = {
+            "dropout" : layer._ratio
+        }
+
+        return output_data
+
+    @classmethod
+    def import_from_dict(cls, layer_data : Dict[ str, Union[ str, int ] ]) -> "Dropout":
+
+        if (cls.DEBUG_MODE):
+            assert isinstance(layer_data, dict)
+
+        dropout_layer = Dropout(
+            ratio = layer_data["dropout"]
+        )
+
+        return dropout_layer
+        
+    def __init__(self, ratio : Union[ int, float ]) -> None:
+
+        super(Dropout, self).__init__()
+        
+        if (self.DEBUG_MODE):
+
+            assert ((isinstance(ratio, int)) or (isinstance(ratio, float)))
+
+            assert ((ratio >= 0) and (ratio <= 1))
+
+        # basic parameters
+
+        self._ratio = ratio
+
+    def forward(self, x : np.ndarray) -> np.ndarray:
+
+        # x : (B, I)
+
+        if (self.DEBUG_MODE):
+            assert isinstance(x, np.ndarray)
+
+        # x : (B, I)
+        self._dropout_mask = np.random.randn(*x.shape) <= (1 - self._ratio)
+
+        # x : (B, I)
+        return x * self._dropout_mask
+
+    def backward(self, gradient : np.ndarray) -> np.ndarray:
+
+        # gradient : (B, O)
+
+        if (self.DEBUG_MODE):
+            assert isinstance(gradient, np.ndarray)
+
+        # gradient : (B, O)
+        return self._dropout_mask * gradient
+
 if (__name__ == "__main__"):
 
     test = 1
